"""
Provide functions to interact with persistent data storage.
"""

import re
import typing
from contextlib import asynccontextmanager
from dataclasses import dataclass
from itertools import product

import asyncpg
import fastapi
import sqlalchemy
from asyncpg.exceptions import UniqueViolationError
from fastapi.exceptions import HTTPException
from loguru import logger
from sqlalchemy import Column, Enum, or_
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import Mapped
from sqlalchemy.sql.expression import Case, ColumnElement, UnaryExpression
from starlette import status
from yarl import URL

from jobbergate_api.config import LogLevelEnum, settings
from jobbergate_api.security import IdentityPayload, PermissionMode, lockdown_with_identity

INTEGRITY_CHECK_EXCEPTIONS = (UniqueViolationError,)


def build_db_url(
    override_db_name: str | None = None,
    force_test: bool = False,
    asynchronous: bool = True,
) -> str:
    """
    Build a database url based on settings.

    If ``force_test`` is set, build from the test database settings.
    If ``asynchronous`` is set, use asyncpg.
    If ``override_db_name`` replace the database name in the settings with the supplied value.
    """
    prefix = "TEST_" if force_test else ""
    db_user = getattr(settings, f"{prefix}DATABASE_USER")
    db_password = getattr(settings, f"{prefix}DATABASE_PSWD")
    db_host = getattr(settings, f"{prefix}DATABASE_HOST")
    db_port = getattr(settings, f"{prefix}DATABASE_PORT")
    db_name = getattr(settings, f"{prefix}DATABASE_NAME") if override_db_name is None else override_db_name
    db_path = "/{}".format(db_name)
    db_scheme = "postgresql+asyncpg" if asynchronous else "postgresql"

    return str(
        URL.build(
            scheme=db_scheme,
            user=db_user,
            password=db_password,
            host=db_host,
            port=db_port,
            path=db_path,
        )
    )


class EngineFactory:
    """
    Provide a factory class that creates engines and keeps track of them in an engine mapping.

    This is used for multi-tenancy and database URL creation at request time.
    """

    engine_map: dict[str, AsyncEngine]

    def __init__(self):
        """
        Initialize the EngineFactory.
        """
        self.engine_map = dict()

    async def cleanup(self):
        """
        Close all engines stored in the engine map and clears the engine_map.
        """
        for engine in self.engine_map.values():
            await engine.dispose()
        self.engine_map = dict()

    def get_engine(self, override_db_name: str | None = None) -> AsyncEngine:
        """
        Get a database engine.

        If the database url is already in the engine map, return the engine stored there. Otherwise, build
        a new one, store it, and return the new engine.
        """
        db_url = build_db_url(
            override_db_name=override_db_name,
            force_test=settings.DEPLOY_ENV.lower() == "test",
        )
        if db_url not in self.engine_map:
            self.engine_map[db_url] = create_async_engine(
                db_url,
                pool_size=settings.DATABASE_POOL_SIZE,
                pool_pre_ping=settings.DATABASE_POOL_PRE_PING,
                max_overflow=settings.DATABASE_POOL_MAX_OVERFLOW,
                logging_name="sqlalchemy.engine",
                echo=settings.LOG_LEVEL == LogLevelEnum.TRACE,
            )
        return self.engine_map[db_url]

    @asynccontextmanager
    async def auto_session(
        self,
        override_db_name: str | None = None,
        commit: bool = True,
    ) -> typing.AsyncIterator[AsyncSession]:
        """
        Get an asynchronous database session.

        Gets a new session from the correct engine in the engine map.
        """
        if settings.DEPLOY_ENV.lower() == "test":
            raise RuntimeError("The auto_session context manager may not be used in unit tests.")

        engine = self.get_engine(override_db_name=override_db_name)
        session = AsyncSession(engine)
        await session.begin()
        try:
            yield session
        except Exception as err:
            logger.warning(f"Rolling back session due to error: {err}")
            await session.rollback()
            raise err
        else:
            if commit is True:
                logger.debug("Committing session")
                await session.commit()
            else:
                logger.debug("Rolling back read-only session")
                await session.rollback()
        finally:
            logger.debug("Closing session")
            await session.close()


engine_factory = EngineFactory()


@dataclass
class SecureSession:
    """
    Provide a container class for an IdentityPayload and AsyncSesson for the current request.
    """

    identity_payload: IdentityPayload
    session: AsyncSession


def secure_session(
    *scopes: str,
    permission_mode: PermissionMode = PermissionMode.SOME,
    commit: bool = True,
    ensure_email: bool = False,
    ensure_organization: bool = False,
    ensure_client_id: bool = False,
):
    """
    Provide an injectable for FastAPI that checks permissions and returns a database session for this request.

    This should be used for all secured routes that need access to the database. It will commit the
    transaction upon completion of the request. If an exception occurs, it will rollback the transaction.
    If multi-tenancy is enabled, it will retrieve a database session for the database associated with the
    client_id found in the requesting user's auth token.

    If testing mode is enabled, it will flush the session instead of committing changes to the database.

    Note that the session should NEVER be explicitly committed anywhere else in the source code.
    """

    async def dependency(
        identity_payload: IdentityPayload = fastapi.Depends(
            lockdown_with_identity(
                *scopes,
                permission_mode=permission_mode,
                ensure_email=ensure_email,
                ensure_organization=ensure_organization,
                ensure_client_id=ensure_client_id,
            )
        ),
    ) -> typing.AsyncIterator[SecureSession]:
        override_db_name = identity_payload.organization_id if settings.MULTI_TENANCY_ENABLED else None
        async with engine_factory.auto_session(override_db_name=override_db_name, commit=commit) as session:
            yield SecureSession(
                identity_payload=identity_payload,
                session=session,
            )

    return dependency


def render_sql(session: AsyncSession, query) -> str:
    """
    Render a sqlalchemy query into a string for debugging.
    """
    return query.compile(dialect=session.bind.dialect, compile_kwargs={"literal_binds": True})


def search_clause(
    search_terms: str,
    searchable_fields: set,
) -> ColumnElement[bool]:
    """
    Create search clause across searchable fields with search terms.

    Regarding the False first argument to or_():
        The or_() function must have one fixed positional argument.
        See: https://docs.sqlalchemy.org/en/20/core/sqlelement.html#sqlalchemy.sql.expression.or_
    """
    search_pairs = product(searchable_fields, search_terms.split())
    search_expressions = (field.ilike(f"%{term}%") for (field, term) in search_pairs)
    return or_(False, *search_expressions)


def _build_enum_sort_clause(sort_column: Column, sort_ascending: bool) -> Case:
    """
    Build a Case statement that can be used as a sort clause for an enum column.

    SQLAlchemy will not sort enums alphabetically by default, but rather by creation order.  Thus, we have to
    force alphabetical sorting using a CASE clause. The logic here is strange. Basically, we have to provide a
    lookup and then a string sort value. If sort order is ascending, each enum value's sort value is itself.
    Otherwise, the sort value is it's opposite in the list. For example::

        Given the enum (GAMMA, ALPHA, DELTA, BETA)
        The default sort order used by SQLAlchemy is (GAMMA, ALPHA, DELTA, BETA)
        Mapping used for ascending sort: {ALPHA: ALPHA, BETA: BETA, DELTA: DELTA, GAMMA: GAMMA}
        Mapping used for descending sort: {ALPHA: GAMMA, BETA: DELTA, DELTA: BETA, GAMMA: ALPHA}

    To understand this more fully, start with this SO question: https://stackoverflow.com/a/23618085/642511

    Note::
        For sorting to work on sqlalchemy.Enum types, the enum _must_ be specified with ``native_enum=False``

        ```
        class SomeTable(Base):
            stuff: Mapped[StuffEnum] = mapped_column(Enum(StuffEnum, native_enum=False))
        ```
    """
    assert isinstance(sort_column.type, Enum)
    sorted_values = sorted(sort_column.type.enums)
    sort_tuple = zip(sorted_values, sorted_values if sort_ascending else reversed(sorted_values))
    return sqlalchemy.case(dict(sort_tuple), value=sort_column)


def sort_clause(
    sort_field: str,
    sortable_fields: set,
    sort_ascending: bool,
) -> typing.Union[Mapped, UnaryExpression, Case]:
    """
    Create a sort clause given a sort field, the list of sortable fields, and a sort_ascending flag.
    """
    sort_column: Mapped[typing.Any] | UnaryExpression | Case | None = None
    for sortable_field in sortable_fields:
        if sortable_field.name == sort_field:
            sort_column = sortable_field
            break

    if sort_column is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid sorting column requested: {sort_field}. Must be one of {sortable_fields}",
        )

    if isinstance(sort_column, Column) and isinstance(sort_column.type, sqlalchemy.Enum):
        sort_column = _build_enum_sort_clause(sort_column, sort_ascending)
    elif not sort_ascending:
        sort_column = sort_column.desc()
    return sort_column


def handle_fk_error(
    _: fastapi.Request,
    err: asyncpg.exceptions.ForeignKeyViolationError,
):
    """
    Unpack metadata from a ForeignKeyViolationError and return a 409 response.
    """
    FK_DETAIL_RX = r"DETAIL:  Key \(id\)=\((?P<pk_id>\d+)\) is still referenced from table \"(?P<table>\w+)\""
    matches = re.search(FK_DETAIL_RX, str(err), re.MULTILINE)
    (table, pk_id) = (None, None)
    if matches:
        table = matches.group("table")
        pk_id = matches.group("pk_id")

    logger.error(f"Delete failed due to foreign-key constraint: {table=} {pk_id=}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_409_CONFLICT,
        content=dict(
            detail=dict(
                message="Delete failed due to foreign-key constraint",
                table=table,
                pk_id=pk_id,
            ),
        ),
    )



"""
Provide a module that describes permissions in the API.
"""

from enum import Enum


class Permissions(str, Enum):
    """
    Describe the permissions that may be used for protecting Jobbergate routes.
    """

    ADMIN = "jobbergate:admin"
    MAINTAINER = "jobbergate:maintainer"
    CLUSTERS_READ = "jobbergate:clusters:read"
    CLUSTERS_UPDATE = "jobbergate:clusters:update"
    JOB_SCRIPTS_CREATE = "jobbergate:job-scripts:create"
    JOB_SCRIPTS_DELETE = "jobbergate:job-scripts:delete"
    JOB_SCRIPTS_READ = "jobbergate:job-scripts:read"
    JOB_SCRIPTS_UPDATE = "jobbergate:job-scripts:update"
    JOB_SUBMISSIONS_CREATE = "jobbergate:job-submissions:create"
    JOB_SUBMISSIONS_DELETE = "jobbergate:job-submissions:delete"
    JOB_SUBMISSIONS_READ = "jobbergate:job-submissions:read"
    JOB_SUBMISSIONS_UPDATE = "jobbergate:job-submissions:update"
    JOB_TEMPLATES_CREATE = "jobbergate:job-templates:create"
    JOB_TEMPLATES_DELETE = "jobbergate:job-templates:delete"
    JOB_TEMPLATES_READ = "jobbergate:job-templates:read"
    JOB_TEMPLATES_UPDATE = "jobbergate:job-templates:update"


def can_bypass_ownership_check(permissions: list[str]) -> bool:
    """
    Determine if the user has permissions that allow them to bypass ownership checks.
    """
    return Permissions.ADMIN in permissions or Permissions.MAINTAINER in permissions



"""
RabbitMQ notification system for Jobbergate.
"""

import asyncio
import json
import socket
from contextlib import asynccontextmanager
from typing import Optional

import aio_pika
from loguru import logger

from jobbergate_api.apps.job_submissions.models import JobSubmission
from jobbergate_api.config import settings


@asynccontextmanager
async def rabbitmq_connect(
    exchange_name=None,
    do_purge=False,
):
    """
    Connect to a RabbitMQ queue and exchange.
    """
    if exchange_name is None:
        exchange_name = settings.RABBITMQ_DEFAULT_EXCHANGE

    event_loop = asyncio.get_running_loop()
    connection = await aio_pika.connect_robust(
        host=settings.RABBITMQ_HOST,
        login=settings.RABBITMQ_USERNAME,
        password=settings.RABBITMQ_PASSWORD,
        client_properties={"connection_name": socket.gethostname()},
        loop=event_loop,
    )
    channel = await connection.channel(publisher_confirms=True)
    exchange = await channel.declare_exchange(
        name=exchange_name,
        type=aio_pika.ExchangeType.DIRECT,
        durable=True,
    )
    queue_name = "jobs"
    routing_key = "status"
    declared_queue = await channel.declare_queue(name=queue_name, durable=True)
    await declared_queue.bind(exchange, routing_key)

    try:
        yield (exchange, declared_queue)

    finally:
        if do_purge:
            await declared_queue.purge(timeout=1)
        await declared_queue.unbind(exchange, routing_key)
        await connection.close()


async def publish_status_change(
    job_submission: JobSubmission,
    organization_id: Optional[str] = None,
):
    """
    Publish a status change for a JobSubmission to the RabbitMQ exchange used for notifications.
    """
    if settings.RABBITMQ_HOST is None:
        return

    logger.debug("Publishing status change to notification queue")

    try:
        message_payload = dict(
            path=f"jobs.job_submissions.{job_submission.id}",
            user_email=job_submission.owner_email,
            action="status",
            additional_context=dict(
                status=job_submission.status,
                slurm_job_state=job_submission.slurm_job_state,
            ),
        )
        async with rabbitmq_connect(exchange_name=organization_id) as (exchange, _):
            await exchange.publish(
                message=aio_pika.Message(
                    body=json.dumps(message_payload).encode("utf-8"),
                    delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
                    headers=dict(organization=organization_id),
                ),
                routing_key="status",
            )
    except Exception as exc:
        logger.error("Failed to publish status change notification")
        logger.exception(exc)



"""
Describe constants for the job_submissions module.
"""

from dataclasses import dataclass

from auto_name_enum import AutoNameEnum, auto
from snick import unwrap


class JobSubmissionStatus(AutoNameEnum):
    """
    Defines the set of possible statuses for a Job Submission.
    """

    CREATED = auto()
    SUBMITTED = auto()
    REJECTED = auto()
    DONE = auto()
    ABORTED = auto()


class SlurmJobState(AutoNameEnum):
    """
    Defines the set of possible states for a job submitted to Slurm.
    """

    BOOT_FAIL = auto()
    CANCELLED = auto()
    COMPLETED = auto()
    CONFIGURING = auto()
    COMPLETING = auto()
    DEADLINE = auto()
    FAILED = auto()
    NODE_FAIL = auto()
    OUT_OF_MEMORY = auto()
    PENDING = auto()
    PREEMPTED = auto()
    RUNNING = auto()
    RESV_DEL_HOLD = auto()
    REQUEUE_FED = auto()
    REQUEUE_HOLD = auto()
    REQUEUED = auto()
    RESIZING = auto()
    REVOKED = auto()
    SIGNALING = auto()
    SPECIAL_EXIT = auto()
    STAGE_OUT = auto()
    STOPPED = auto()
    SUSPENDED = auto()
    TIMEOUT = auto()

    # Special job state indicating that something went wrong with SLURM and the job state cannot be recovered
    UNKNOWN = auto()


@dataclass
class SlurmJobStateDetails:
    """
    Defines the details for a given SlurmJobState including abbreviation and description.
    """

    abbreviation: str
    description: str
    is_abort_status: bool = False
    is_done_status: bool = False


slurm_job_state_details = {
    SlurmJobState.BOOT_FAIL: SlurmJobStateDetails(
        abbreviation="BF",
        description=unwrap(
            """
            Job terminated due to launch failure, typically due to a hardware failure
            (e.g. unable to boot the node or block and the job can not be requeued).
            """
        ),
        is_abort_status=True,
    ),
    SlurmJobState.CANCELLED: SlurmJobStateDetails(
        abbreviation="CA",
        description=unwrap(
            """
            Job was explicitly cancelled by the user or system administrator.
            The job may or may not have been initiated.
            """
        ),
        is_abort_status=True,
    ),
    SlurmJobState.COMPLETED: SlurmJobStateDetails(
        abbreviation="CD",
        description="Job has terminated all processes on all nodes with an exit code of zero.",
        is_done_status=True,
    ),
    SlurmJobState.CONFIGURING: SlurmJobStateDetails(
        abbreviation="CF",
        description=unwrap(
            """
            Job has been allocated resources, but are waiting for them to become ready for use
            (e.g. booting).
            """
        ),
    ),
    SlurmJobState.COMPLETING: SlurmJobStateDetails(
        abbreviation="CG",
        description="Job is in the process of completing. Some processes on some nodes may still be active.",
    ),
    SlurmJobState.DEADLINE: SlurmJobStateDetails(
        abbreviation="DL",
        description="Job terminated on deadline.",
        is_abort_status=True,
    ),
    SlurmJobState.FAILED: SlurmJobStateDetails(
        abbreviation="F",
        description="Job terminated with non-zero exit code or other failure condition.",
        is_abort_status=True,
    ),
    SlurmJobState.NODE_FAIL: SlurmJobStateDetails(
        abbreviation="NF",
        description="Job terminated due to failure of one or more allocated nodes.",
        is_abort_status=True,
    ),
    SlurmJobState.OUT_OF_MEMORY: SlurmJobStateDetails(
        abbreviation="OOM",
        description="Job experienced out of memory error.",
    ),
    SlurmJobState.PENDING: SlurmJobStateDetails(
        abbreviation="PD",
        description="Job is awaiting resource allocation.",
    ),
    SlurmJobState.PREEMPTED: SlurmJobStateDetails(
        abbreviation="PR",
        description="Job terminated due to preemption.",
        is_abort_status=True,
    ),
    SlurmJobState.RUNNING: SlurmJobStateDetails(
        abbreviation="R",
        description="Job currently has an allocation.",
    ),
    SlurmJobState.RESV_DEL_HOLD: SlurmJobStateDetails(
        abbreviation="RD",
        description="Job is being held after requested reservation was deleted.",
    ),
    SlurmJobState.REQUEUE_FED: SlurmJobStateDetails(
        abbreviation="RF",
        description="Job is being requeued by a federation.",
    ),
    SlurmJobState.REQUEUE_HOLD: SlurmJobStateDetails(
        abbreviation="RH",
        description="Held job is being requeued.",
    ),
    SlurmJobState.REQUEUED: SlurmJobStateDetails(
        abbreviation="RQ",
        description="Completing job is being requeued.",
    ),
    SlurmJobState.RESIZING: SlurmJobStateDetails(
        abbreviation="RS",
        description="Job is about to change size.",
    ),
    SlurmJobState.REVOKED: SlurmJobStateDetails(
        abbreviation="RV",
        description="Sibling was removed from cluster due to other cluster starting the job.",
    ),
    SlurmJobState.SIGNALING: SlurmJobStateDetails(
        abbreviation="SI",
        description="Job is being signaled.",
    ),
    SlurmJobState.SPECIAL_EXIT: SlurmJobStateDetails(
        abbreviation="SE",
        description=unwrap(
            """
            The job was requeued in a special state.
            This state can be set by users, typically in EpilogSlurmctld,
            if the job has terminated with a particular exit value.
            """
        ),
        is_abort_status=True,
    ),
    SlurmJobState.STAGE_OUT: SlurmJobStateDetails(
        abbreviation="SO",
        description="Job is staging out files.",
    ),
    SlurmJobState.STOPPED: SlurmJobStateDetails(
        abbreviation="ST",
        description=unwrap(
            """
            Job has an allocation, but execution has been stopped with SIGSTOP signal.
            CPUS have been retained by this job.
            """
        ),
    ),
    SlurmJobState.SUSPENDED: SlurmJobStateDetails(
        abbreviation="S",
        description=unwrap(
            """
            Job has an allocation, but execution has been suspended
            and CPUs have been released for other jobs.
            """
        ),
    ),
    SlurmJobState.TIMEOUT: SlurmJobStateDetails(
        abbreviation="TO",
        description="Job terminated upon reaching its time limit.",
        is_abort_status=True,
    ),
    SlurmJobState.UNKNOWN: SlurmJobStateDetails(
        abbreviation="UK",
        description="Indicates that something went wrong with SLURM and the job state cannot be recovered",
        is_abort_status=True,
    ),
}



"""Functionalities to be shared by all models."""

from typing import cast

from inflection import tableize
from pendulum.datetime import DateTime as PendulumDateTime
from snick import conjoin
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import DeclarativeBase, Mapped, declared_attr, mapped_column
from sqlalchemy.sql.expression import Select


class Base(DeclarativeBase):
    """
    Base class for all models.

    References:
        https://docs.sqlalchemy.org/en/20/orm/declarative_mixins.html
    """


class CommonMixin:
    """
    Provide a dynamic table and helper methods for displaying instances.
    """

    @declared_attr.directive
    @classmethod
    def __tablename__(cls) -> str:
        """
        Dynamically create table name based on the class name.
        """
        return tableize(cls.__name__)

    def _iter_cols(self):
        """
        Iterate over the columns of the class.
        """
        for col in inspect(self.__class__).columns.keys():
            yield (col, getattr(self, col))

    def __str__(self):
        """
        Produce a pretty string representation of the class instance.
        """
        primary_keys = [pk.name for pk in inspect(self.__class__).primary_key]
        primary_key_str = ", ".join([f"{pk}: {getattr(self, pk)}" for pk in primary_keys])
        return conjoin(
            f"{self.__class__.__name__}: ({primary_key_str})",
            *[f"{k}: {v}" for (k, v) in self._iter_cols() if k not in primary_keys],
            join_str="\n  ",
        )


class IdMixin:
    """
    Provide an id primary_key column.

    Attributes:
        id: The id of the job script template.
        cloned_from: Specify the id of the row that this row was cloned from.
    """

    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    @declared_attr
    def cloned_from_id(cls) -> Mapped[int | None]:
        """
        Dynamically create a cloned_from_id column.
        """
        return mapped_column(
            Integer,
            ForeignKey(cast(Column[int], cls.id), name="clone", ondelete="SET NULL"),
            nullable=True,
        )


class TimestampMixin:
    """
    Add timestamp columns to a table.

    Attributes:
        created_at: The date and time when the job script template was created.
        updated_at: The date and time when the job script template was updated.
    """

    created_at: Mapped[PendulumDateTime] = mapped_column(
        DateTime(timezone=True), nullable=False, default=PendulumDateTime.utcnow
    )
    updated_at: Mapped[PendulumDateTime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        default=PendulumDateTime.utcnow,
        onupdate=PendulumDateTime.utcnow,
    )


class OwnerMixin:
    """
    Add an owner email columns to a table.

    Attributes:
        owner_email: The email of the owner of the job script template.
    """

    owner_email: Mapped[str] = mapped_column(String, nullable=False, index=True)


class NameMixin:
    """
    Add name and description columns to a table.

    Attributes:
        name: The name of the job script template.
        description: The description of the job script template.
    """

    name: Mapped[str] = mapped_column(String, nullable=False, index=True)
    description: Mapped[str | None] = mapped_column(String, default="")


class ArchiveMixin:
    """
    Add is_archived column to a table.

    Attributes:
        is_archived: Specify is a row is considered archived, hidden it by default when listing rows.
    """

    is_archived: Mapped[bool] = mapped_column(Boolean, nullable=False, default=False)


class CrudMixin(CommonMixin, IdMixin, TimestampMixin, OwnerMixin, NameMixin, ArchiveMixin):
    """
    Add needed columns and declared attributes for all models that support a CrudService.
    """

    @classmethod
    def searchable_fields(cls):
        """
        Describe the fields that may be used in search queries.
        """
        return {
            cls.name,
            cls.description,
            cls.owner_email,
        }

    @classmethod
    def sortable_fields(cls):
        """
        Describe the fields that may be used for sorting queries.
        """
        return {
            cls.id,
            cls.name,
            cls.owner_email,
            cls.created_at,
            cls.updated_at,
        }

    @classmethod
    def include_files(cls, query: Select) -> Select:
        """
        Include custom options on a query to eager load files.

        This should be overridden by derived classes.
        """
        raise NotImplementedError(
            f"Derived classes should override include_files. {cls.__tablename__} does not include it."
        )

    @classmethod
    def include_parent(cls, query: Select) -> Select:
        """
        Include custom options on a query to eager load parent data.

        This should be overridden by derived classes.
        """
        raise NotImplementedError(
            f"Derived classes should override include_parent. {cls.__tablename__} does not include it."
        )


class FileMixin(CommonMixin, TimestampMixin):
    """
    Add needed columns and declared attributes for all models that support a FileService.

    Attributes:
        parent_id:   The id of the parent row in another table.
                     Note: Derived classes should override this attribute to make it a foreign key as well.
        description: The description of the job script template.
    """

    parent_id: Mapped[int] = mapped_column(Integer, primary_key=True)
    filename: Mapped[str] = mapped_column(String, primary_key=True)

    @hybrid_property
    def file_key(self) -> str:
        """
        Dynamically define the s3 key for the file.
        """
        return f"{self.__tablename__}/{self.parent_id}/{self.filename}"



"""
Email notification system for Jobbergate.
"""

from dataclasses import dataclass
from typing import List, Optional, Union

from buzz import Buzz
from loguru import logger
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail

from jobbergate_api.config import settings


class EmailNotificationError(Buzz):
    """
    Custom error to be raised for problems at the email notification system.
    """

    pass


@dataclass
class EmailManager:
    """
    Email manager.
    """

    email_client: SendGridAPIClient
    from_email: Optional[str] = ""

    def send_email(
        self, to_emails: Union[str, List[str]], subject: str, skip_on_failure: bool = False, **kwargs
    ) -> None:
        """
        Send an email using this manager.
        """
        with EmailNotificationError.handle_errors(
            f"Error while sending email from_email={self.from_email}, {to_emails=}",
            do_except=lambda params: logger.warning(params.final_message),
            re_raise=not skip_on_failure,
        ):
            message = self._build_message(to_emails, subject, **kwargs)
            self.email_client.send(message)

    def _build_message(self, to_emails: Union[str, List[str]], subject: str, **kwargs) -> Mail:
        """
        Build an email message.
        """
        EmailNotificationError.require_condition(self.from_email, "The value from_email is empty.")

        with EmailNotificationError.handle_errors(
            "Error while creating the message",
            do_except=lambda params: logger.error(params.final_message),
        ):
            message = Mail(
                from_email=self.from_email,
                to_emails=to_emails,
                subject=subject,
                **kwargs,
            )
        return message


def notify_submission_rejected(
    job_submission_id: Union[str, int], report_message: str, to_emails: Union[str, List[str]]
) -> None:
    """
    Notify an email or a list of emails about a job submission that has been rejected.
    """
    subject = f"Job Submission Rejected (id={job_submission_id})"

    logger.debug(f"Notifying {to_emails=} that {job_submission_id=} was rejected: {report_message=}")
    email_manager.send_email(to_emails, subject, skip_on_failure=True, plain_text_content=report_message)


email_client = SendGridAPIClient(settings.SENDGRID_API_KEY)
email_manager = EmailManager(
    email_client,
    settings.SENDGRID_FROM_EMAIL,
)
